#define _GNU_SOURCE

#include <errno.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <time.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <pwd.h>
#include <grp.h>


#include <config.h>
#include <Conn.h>
#include <ghostpop3_conf.h>


/* Global variables */
static unsigned short	debug = 0;

static FILE		*Logf = NULL;
static char		*conf_file = GHOSTPOP3_ETC_DIR "/ghostpop3.conf";
static int		port = 110;
static unsigned short	ipv4 = 1, ipv6 = 1;
static unsigned int	maxconn = 5000;
static unsigned short	log_user_pass = 1;

struct priv {
	char user[128];
};

static void Log(unsigned short level, char *format, ...)
{
	va_list	ap;

	if (level > debug)
		return;

	va_start(ap, format);
	vfprintf(Logf, format, ap);
	va_end(ap);
}

static void ghostpop3_accept(struct Conn *C)
{
	char *s;

	Log(4, "%s (A connection was accepted from [%s] on slot %d. Set POLLIN and enq greeting...\n",
		__FUNCTION__, C->addr, C->slot);

	s = "+OK ghostpop3 server - (C)opyright Catalin BOIE - http://kernel.umbrella.ro/\r\n";
	Conn_enqueue(C, s, strlen(s));

	C->private = malloc(sizeof(struct priv));
	if (!C->private)
		Log(0, "\tCannot alloc priv structure!\n");
}

static void ghostpop3_close(struct Conn *C)
{
	Log(5, "%s (A connection will be closed [%s] on slot %d)\n",
		__FUNCTION__, C->addr, C->slot);

	if (C->private)
		free(C->private);
}

static void ghostpop3_data(struct Conn *C)
{
	char	*p, *dump;
	char	buf[4096];
	struct priv *priv = C->private;
	unsigned int do_send = 1;

	if (debug >= 8) {
		dump = Conn_dump(C->ibuf + C->ibuf_head, C->ibuf_tail - C->ibuf_head);
		Log(0, "%s: recv head=%d tail=%d bytes on slot %d: [%s]\n",
			__FUNCTION__, C->ibuf_head, C->ibuf_tail,
			C->slot, dump);
		free(dump);
	}

	p = C->ibuf + C->ibuf_head;
	if (strncasecmp(p, "RETR", 4) == 0) {
		strcpy(buf, "-ERR\r\n");
	} else if (strncasecmp(p, "LIST", 4) == 0) {
		strcpy(buf, "+OK 0 messages (0 octets)\r\n.\r\n");
	} else if (strncasecmp(p, "STAT", 4) == 0) {
		strcpy(buf, "+OK 0 0\r\n");
	} else if (strncasecmp(p, "QUIT", 4) == 0) {
		Conn_close(C);
		do_send = 0;
	} else if (strncasecmp(p, "USER ", 5) == 0) {
		if (priv)
			snprintf(priv->user, sizeof(priv->user), "%s", p + 5);
		strcpy(buf, "+OK\r\n");
	} else if (strncasecmp(p, "PASS ", 5) == 0) {
		if (priv && (log_user_pass == 1))
			Log(0, "USER/PASS: [%s]-[%s]\n",
				fix(priv->user), fix(p + 5));
		strcpy(buf, "+OK\r\n");
	} else {
		strcpy(buf, "-ERR\r\n");
	}

	if (do_send == 1) {
		Log(8, "\tSend back: %s", buf);
		Conn_enqueue(C, buf, strlen(buf));
	}

	Conn_eatall(C);
}

static void ghostpop3_error(struct Conn *C)
{
	Log(1, "%s id=%llu\n",
		__FUNCTION__, Conn_getid(C));
}

int main(void)
{
	struct Conn *I4 = NULL, *I6 = NULL;
	int ret;
	struct timeval start;

	/* Set umask so logfile will not be readable */
	umask(S_IRUSR | S_IWUSR);

	ret = ghostpop3_conf_load(conf_file);
	if (ret != 0) {
		fprintf(stderr, "Cannot load conf file [%s] [%s]!\n",
			conf_file, ghostpop3_conf_strerror());
		return 1;
	}

	if (strcmp(conf_daemon, "on") == 0)
		daemon(0, 0);

	port = strtol(conf_port, NULL, 10);
	if (strncmp(conf_ipv4, "of", 2) == 0)
		ipv4 = 0;
	if (strncmp(conf_ipv6, "of", 2) == 0)
		ipv6 = 0;
	maxconn = strtol(conf_maxconn, NULL, 10);
	debug = strtoul(conf_debug, NULL, 10);
	if (strncmp(conf_log_user_pass, "of", 2) == 0)
		log_user_pass = 0;

	Logf = fopen(conf_log, "a+");
	if (!Logf) {
		fprintf(stderr, "Cannot open log file [%s] [%s]\n",
			conf_log, strerror(errno));
		return 1;
	}
	setlinebuf(Logf);

	/* switch to other user */
	if (geteuid() == 0) {
		struct passwd *pw;
		struct group *g;
		uid_t uid;
		gid_t gid;

		/* switch group */
		g = getgrnam(conf_group);
		if (!g) {
			fprintf(stderr, "Cannot get info about group [%s] [%s]!\n",
				conf_group, strerror(errno));
			return 1;
		}
		gid = g->gr_gid;
		ret = setresgid(gid, gid, gid);
		if (ret != 0) {
			fprintf(stderr, "Cannot switch to group [%s] [%s]!\n",
				conf_group, strerror(errno));
			return 1;
		}

		/* switch user */
		pw = getpwnam(conf_user);
		if (!pw) {
			fprintf(stderr, "Cannot get info about user [%s] [%s]!\n",
				conf_user, strerror(errno));
			return 1;
		}
		uid = pw->pw_uid;
		ret = setresuid(uid, uid, uid);
		if (ret != 0) {
			fprintf(stderr, "Cannot switch to user [%s] [%s]!\n",
				conf_user, strerror(errno));
			return 1;
		}
	}

	gettimeofday(&start, NULL);
	Log(0, "\nStarting at %ld...\n", start.tv_sec);
	Log(0, "\tPort=%d\n", port);
	Log(0, "\tdebug=%d\n", debug);
	Log(0, "\tipv4=%s ipv6=%s\n", (ipv4 == 1) ? "on" : "off", (ipv6 == 1) ? "on" : "off");
	Log(0, "\tMessage directory: %s\n", conf_msgdir);
	Log(0, "\tlog_user_pass = %s\n", log_user_pass == 1 ? "on" : "off");
	Log(0, "\tMaximum nuber of connections allowed=%d\n", maxconn);

	Conn_debug(Logf, debug);

	ret = Conn_init(maxconn);
	if (ret == -1) {
		Log(0, "%s", Conn_strerror());
		return 1;
	}

	if (ipv4 == 1) {
		Log(9, "Try to register IPv4 socket...\n");
		I4 = Conn_socket(PF_INET, SOCK_STREAM, port);
		if (!I4) {
			Log(1, "Cannot bind on ipv4 socket [%s].\n",
				Conn_strerror());
			return 1;
		}
	}

	if (ipv6 == 1) {
		Log(9, "Try to register IPv6 socket...\n");
		I6 = Conn_socket(PF_INET6, SOCK_STREAM, port);
		if (!I6) {
			Log(1, "Cannot bind on ipv6 socket [%s].\n",
				Conn_strerror());
			return 1;
		}
	}

	Conn_accept_cb = ghostpop3_accept;
	Conn_data_cb = ghostpop3_data;
	Conn_close_cb = ghostpop3_close;
	Conn_error_cb = ghostpop3_error;

	while (1) {
		ret = Conn_poll(10000);
		if (ret == -1) {
			Log(0, "Error calling Conn_poll [%s]!\n",
				Conn_strerror());
			return 1;
		}

		if (ret == 0)
			break;
	}

	Log(0, "Finish!\n\n");

	return 0;
}
